# Authors:
#   Jason Gerard DeRose <jderose@redhat.com>
#
# Copyright (C) 2008-2016  Red Hat
# see file 'COPYING' for use and warranty inmsgion
#
# This program is free software; you can redistribute it and/or modify
# it under the terms of the GNU General Public License as published by
# the Free Software Foundation, either version 3 of the License, or
# (at your option) any later version.
#
# This program is distributed in the hope that it will be useful,
# but WITHOUT ANY WARRANTY; without even the implied warranty of
# MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
# GNU General Public License for more details.
#
# You should have received a copy of the GNU General Public License
# along with this program.  If not, see <http://www.gnu.org/licenses/>.

"""
Custom exception classes (some which are RPC transparent).

`PrivateError` and its subclasses are custom IPA excetions that will *never* be
forwarded in a Remote Procedure Call (RPC) response.

On the other hand, `PublicError` and its subclasses can be forwarded in an RPC
response.  These public errors each carry a unique integer error code as well as
a gettext translated error message (translated at the time the exception is
raised).  The purpose of the public errors is to relay information about
*expected* user errors, service availability errors, and so on.  They should
*never* be used for *unexpected* programmatic or run-time errors.

For security reasons it is *extremely* important that arbitrary exceptions *not*
be forwarded in an RPC response.  Unexpected exceptions can easily contain
compromising information in their error messages.  Any time the server catches
any exception that isn't a `PublicError` subclass, it should raise an
`InternalError`, which itself always has the same, static error message (and
therefore cannot be populated with information about the true exception).

The public errors are arranging into five main blocks of error code ranges:

    =============  ========================================
     Error codes                 Exceptions
    =============  ========================================
    1000 - 1999    `AuthenticationError` and its subclasses
    2000 - 2999    `AuthorizationError` and its subclasses
    3000 - 3999    `InvocationError` and its subclasses
    4000 - 4999    `ExecutionError` and its subclasses
    5000 - 5999    `GenericError` and its subclasses
    =============  ========================================

Within these five blocks some sub-ranges are already allocated for certain types
of error messages, while others are reserved for future use.  Here are the
current block assignments:

    - **900-5999** `PublicError` and its subclasses

        - **901 - 907**  Assigned to special top-level public errors

        - **908 - 999**  *Reserved for future use*

        - **1000 - 1999**  `AuthenticationError` and its subclasses

            - **1001 - 1099**  Open for general authentication errors

            - **1100 - 1199**  `KerberosError` and its subclasses

            - **1200 - 1299**  `SessionError` and its subclasses

            - **1300 - 1999**  *Reserved for future use*

        - **2000 - 2999**  `AuthorizationError` and its subclasses

            - **2001 - 2099**  Open for general authorization errors

            - **2100 - 2199**  `ACIError` and its subclasses

            - **2200 - 2999**  *Reserved for future use*

        - **3000 - 3999**  `InvocationError` and its subclasses

            - **3001 - 3099**  Open for general invocation errors

            - **3100 - 3199**  *Reserved for future use*

        - **4000 - 4999**  `ExecutionError` and its subclasses

            - **4001 - 4099**  Open for general execution errors

            - **4100 - 4199**  `BuiltinError` and its subclasses

            - **4200 - 4299**  `LDAPError` and its subclasses

            - **4300 - 4399**  `CertificateError` and its subclasses

            - **4400 - 4499**  `DNSError` and (some of) its subclasses

            - **4500 - 4999**  *Reserved for future use*

        - **5000 - 5999**  `GenericError` and its subclasses

            - **5001 - 5099**  Open for generic errors

            - **5100 - 5999**  *Reserved for future use*
"""

import six

from ipalib.text import ngettext as ungettext
from ipalib import messages


class PrivateError(Exception):
    """
    Base class for exceptions that are *never* forwarded in an RPC response.
    """

    format = ''

    def __init__(self, **kw):
        self.msg = self.format % kw
        self.kw = kw
        for (key, value) in kw.items():
            assert not hasattr(self, key), 'conflicting kwarg %s.%s = %r' % (
                self.__class__.__name__, key, value,
            )
            setattr(self, key, value)
        Exception.__init__(self, self.msg)

    if six.PY3:
        @property
        def message(self):
            return str(self)


class SubprocessError(PrivateError):
    """
    Raised when ``subprocess.call()`` returns a non-zero exit status.

    This custom exception is needed because Python 2.4 doesn't have the
    ``subprocess.CalledProcessError`` exception (which was added in Python 2.5).

    For example:

    >>> raise SubprocessError(returncode=2, argv=('ls', '-lh', '/no-foo/'))
    Traceback (most recent call last):
      ...
    SubprocessError: return code 2 from ('ls', '-lh', '/no-foo/')

    The exit code of the sub-process is available via the ``returncode``
    instance attribute.  For example:

    >>> e = SubprocessError(returncode=1, argv=('/bin/false',))
    >>> e.returncode
    1
    >>> e.argv  # argv is also available
    ('/bin/false',)
    """

    format = 'return code %(returncode)d from %(argv)r'


class PluginSubclassError(PrivateError):
    """
    Raised when a plugin doesn't subclass from an allowed base.

    For example:

    >>> raise PluginSubclassError(plugin='bad', bases=('base1', 'base2'))
    Traceback (most recent call last):
      ...
    PluginSubclassError: 'bad' not subclass of any base in ('base1', 'base2')

    """

    format = '%(plugin)r not subclass of any base in %(bases)r'


class PluginDuplicateError(PrivateError):
    """
    Raised when the same plugin class is registered more than once.

    For example:

    >>> raise PluginDuplicateError(plugin='my_plugin')
    Traceback (most recent call last):
      ...
    PluginDuplicateError: 'my_plugin' was already registered
    """

    format = '%(plugin)r was already registered'


class PluginOverrideError(PrivateError):
    """
    Raised when a plugin overrides another without using ``override=True``.

    For example:

    >>> raise PluginOverrideError(base='Command', name='env', plugin='my_env')
    Traceback (most recent call last):
      ...
    PluginOverrideError: unexpected override of Command.env with 'my_env'
    """

    format = 'unexpected override of %(base)s.%(name)s with %(plugin)r'


class PluginMissingOverrideError(PrivateError):
    """
    Raised when a plugin overrides another that has not been registered.

    For example:

    >>> raise PluginMissingOverrideError(base='Command', name='env', plugin='my_env')
    Traceback (most recent call last):
      ...
    PluginMissingOverrideError: Command.env not registered, cannot override with 'my_env'
    """

    format = '%(base)s.%(name)s not registered, cannot override with %(plugin)r'


class SkipPluginModule(PrivateError):
    """
    Raised to abort the loading of a plugin module.
    """

    format = '%(reason)s'


class PluginsPackageError(PrivateError):
    """
    Raised when ``package.plugins`` is a module instead of a sub-package.
    """

    format = '%(name)s must be sub-package, not module: %(file)r'


class PluginModuleError(PrivateError):
    """
    Raised when a module is not a valid plugin module.
    """

    format = '%(name)s is not a valid plugin module'


##############################################################################
# Public errors:

_texts = []

def _(message):
    _texts.append(message)
    return message


class PublicError(Exception):
    """
    **900** Base class for exceptions that can be forwarded in an RPC response.
    """
    def __init__(self, format=None, message=None, **kw):
        messages.process_message_arguments(self, format, message, **kw)
        super(PublicError, self).__init__(self.msg)

    errno = 900
    rval = 1
    format = None

    if six.PY3:
        @property
        def message(self):
            return str(self)


class VersionError(PublicError):
    """
    **901** Raised when client and server versions are incompatible.

    For example:

    >>> raise VersionError(cver='2.0', sver='2.1', server='https://localhost')
    Traceback (most recent call last):
      ...
    VersionError: 2.0 client incompatible with 2.1 server at 'https://localhost'

    """

    errno = 901
    format = _("%(cver)s client incompatible with %(sver)s server at '%(server)s'")


class UnknownError(PublicError):
    """
    **902** Raised when client does not know error it caught from server.

    For example:

    >>> raise UnknownError(code=57, server='localhost', error=u'a new error')
    ...
    Traceback (most recent call last):
      ...
    UnknownError: unknown error 57 from localhost: a new error

    """

    errno = 902
    format = _('unknown error %(code)d from %(server)s: %(error)s')


class InternalError(PublicError):
    """
    **903** Raised to conceal a non-public exception.

    For example:

    >>> raise InternalError()
    Traceback (most recent call last):
      ...
    InternalError: an internal error has occurred
    """

    errno = 903
    format = _('an internal error has occurred')

    def __init__(self, message=None):
        """
        Security issue: ignore any information given to constructor.
        """
        PublicError.__init__(self)


class ServerInternalError(PublicError):
    """
    **904** Raised when client catches an `InternalError` from server.

    For example:

    >>> raise ServerInternalError(server='https://localhost')
    Traceback (most recent call last):
      ...
    ServerInternalError: an internal error has occurred on server at 'https://localhost'
    """

    errno = 904
    format = _("an internal error has occurred on server at '%(server)s'")


class CommandError(PublicError):
    """
    **905** Raised when an unknown command is called.

    For example:

    >>> raise CommandError(name='foobar')
    Traceback (most recent call last):
      ...
    CommandError: unknown command 'foobar'
    """

    errno = 905
    format = _("unknown command '%(name)s'")


class ServerCommandError(PublicError):
    """
    **906** Raised when client catches a `CommandError` from server.

    For example:

    >>> e = CommandError(name='foobar')
    >>> raise ServerCommandError(error=str(e), server='https://localhost')
    Traceback (most recent call last):
      ...
    ServerCommandError: error on server 'https://localhost': unknown command 'foobar'
    """

    errno = 906
    format = _("error on server '%(server)s': %(error)s")


class NetworkError(PublicError):
    """
    **907** Raised when a network connection cannot be created.

    For example:

    >>> raise NetworkError(uri='ldap://localhost:389', error=_(u'Connection refused'))
    Traceback (most recent call last):
      ...
    NetworkError: cannot connect to 'ldap://localhost:389': Connection refused
    """

    errno = 907
    format = _("cannot connect to '%(uri)s': %(error)s")


class ServerNetworkError(PublicError):
    """
    **908** Raised when client catches a `NetworkError` from server.
    """

    errno = 908
    format = _("error on server '%(server)s': %(error)s")


class JSONError(PublicError):
    """
    **909** Raised when server received a malformed JSON-RPC request.
    """

    errno = 909
    format = _('Invalid JSON-RPC request: %(error)s')


class XMLRPCMarshallError(PublicError):
    """
    **910** Raised when the XML-RPC lib cannot marshall the request

    For example:

    >>> raise XMLRPCMarshallError(error=_('int exceeds XML-RPC limits'))
    Traceback (most recent call last):
      ...
    XMLRPCMarshallError: error marshalling data for XML-RPC transport: int exceeds XML-RPC limits
    """

    errno = 910
    format = _('error marshalling data for XML-RPC transport: %(error)s')


class RefererError(PublicError):
    """
    **911** Raised when the request does not contain an HTTP referer

    For example:

    >>> raise RefererError(referer='referer')
    Traceback (most recent call last):
      ...
    RefererError: Missing or invalid HTTP Referer, referer
    """

    errno = 911
    format = _('Missing or invalid HTTP Referer, %(referer)s')


class EnvironmentError(PublicError):
    """
    **912** Raised when a command is called with invalid environment settings
    """

    errno = 912


##############################################################################
# 1000 - 1999: Authentication errors
class AuthenticationError(PublicError):
    """
    **1000** Base class for authentication errors (*1000 - 1999*).
    """

    errno = 1000


class KerberosError(AuthenticationError):
    """
    **1100** Base class for Kerberos authentication errors (*1100 - 1199*).

    For example:

    >>> raise KerberosError(major=_('Unspecified GSS failure.  Minor code may provide more information'), minor=_('No credentials cache found'))
    Traceback (most recent call last):
      ...
    KerberosError: Kerberos error: Unspecified GSS failure.  Minor code may provide more information/No credentials cache found

    """

    errno = 1100
    format= _('Kerberos error: %(major)s/%(minor)s')


class CCacheError(KerberosError):
    """
    **1101** Raised when sever does not receive Kerberose credentials.

    For example:

    >>> raise CCacheError()
    Traceback (most recent call last):
      ...
    CCacheError: did not receive Kerberos credentials

    """

    errno = 1101
    format = _('did not receive Kerberos credentials')


class ServiceError(KerberosError):
    """
    **1102** Raised when service is not found in Kerberos DB.

    For example:

    >>> raise ServiceError(service='HTTP@localhost')
    Traceback (most recent call last):
      ...
    ServiceError: Service 'HTTP@localhost' not found in Kerberos database
    """

    errno = 1102
    format = _("Service '%(service)s' not found in Kerberos database")


class NoCCacheError(KerberosError):
    """
    **1103** Raised when a client attempts to use Kerberos without a ccache.

    For example:

    >>> raise NoCCacheError()
    Traceback (most recent call last):
      ...
    NoCCacheError: No credentials cache found
    """

    errno = 1103
    format = _('No credentials cache found')


class TicketExpired(KerberosError):
    """
    **1104** Raised when a client attempts to use an expired ticket

    For example:

    >>> raise TicketExpired()
    Traceback (most recent call last):
      ...
    TicketExpired: Ticket expired
    """

    errno = 1104
    format = _('Ticket expired')


class BadCCachePerms(KerberosError):
    """
    **1105** Raised when a client has bad permissions on their ccache

    For example:

    >>> raise BadCCachePerms()
    Traceback (most recent call last):
      ...
    BadCCachePerms: Credentials cache permissions incorrect
    """

    errno = 1105
    format = _('Credentials cache permissions incorrect')


class BadCCacheFormat(KerberosError):
    """
    **1106** Raised when a client has a misformated ccache

    For example:

    >>> raise BadCCacheFormat()
    Traceback (most recent call last):
      ...
    BadCCacheFormat: Bad format in credentials cache
    """

    errno = 1106
    format = _('Bad format in credentials cache')


class CannotResolveKDC(KerberosError):
    """
    **1107** Raised when the KDC can't be resolved

    For example:

    >>> raise CannotResolveKDC()
    Traceback (most recent call last):
      ...
    CannotResolveKDC: Cannot resolve KDC for requested realm
    """

    errno = 1107
    format = _('Cannot resolve KDC for requested realm')


class SessionError(AuthenticationError):
    """
    **1200** Base class for Session errors (*1200 - 1299*).

    For example:

    """

    errno = 1200
    format= _('Session error')


class InvalidSessionPassword(SessionError):
    """
    **1201** Raised when we cannot obtain a TGT for a principal.
    """
    errno = 1201
    format= _('Principal %(principal)s cannot be authenticated: %(message)s')

class PasswordExpired(InvalidSessionPassword):
    """
    **1202** Raised when we cannot obtain a TGT for a principal because the password is expired.
    """
    errno = 1202

class KrbPrincipalExpired(SessionError):
    """
    **1203** Raised when Kerberos Principal is expired.
    """
    errno = 1203

class UserLocked(SessionError):
    """
    **1204** Raised when a user account is locked.
    """
    errno = 1204

##############################################################################
# 2000 - 2999: Authorization errors
class AuthorizationError(PublicError):
    """
    **2000** Base class for authorization errors (*2000 - 2999*).
    """

    errno = 2000


class ACIError(AuthorizationError):
    """
    **2100** Base class for ACI authorization errors (*2100 - 2199*).
    """

    errno = 2100
    format = _('Insufficient access: %(info)s')



##############################################################################
# 3000 - 3999: Invocation errors

class InvocationError(PublicError):
    """
    **3000** Base class for command invocation errors (*3000 - 3999*).
    """

    errno = 3000


class EncodingError(InvocationError):
    """
    **3001** Raised when received text is incorrectly encoded.
    """

    errno = 3001


class BinaryEncodingError(InvocationError):
    """
    **3002** Raised when received binary data is incorrectly encoded.
    """

    errno = 3002


class ZeroArgumentError(InvocationError):
    """
    **3003** Raised when a command is called with arguments but takes none.

    For example:

    >>> raise ZeroArgumentError(name='ping')
    Traceback (most recent call last):
      ...
    ZeroArgumentError: command 'ping' takes no arguments
    """

    errno = 3003
    format = _("command '%(name)s' takes no arguments")


class MaxArgumentError(InvocationError):
    """
    **3004** Raised when a command is called with too many arguments.

    For example:

    >>> raise MaxArgumentError(name='user_add', count=2)
    Traceback (most recent call last):
      ...
    MaxArgumentError: command 'user_add' takes at most 2 arguments
    """

    errno = 3004

    def __init__(self, message=None, **kw):
        if message is None:
            format = ungettext(
                "command '%(name)s' takes at most %(count)d argument",
                "command '%(name)s' takes at most %(count)d arguments",
                kw['count']
            )
        else:
            format = None
        InvocationError.__init__(self, format, message, **kw)


class OptionError(InvocationError):
    """
    **3005** Raised when a command is called with unknown options.
    """

    errno = 3005


class OverlapError(InvocationError):
    """
    **3006** Raised when arguments and options overlap.

    For example:

    >>> raise OverlapError(names=['givenname', 'login'])
    Traceback (most recent call last):
      ...
    OverlapError: overlapping arguments and options: ['givenname', 'login']
    """

    errno = 3006
    format = _("overlapping arguments and options: %(names)s")


class RequirementError(InvocationError):
    """
    **3007** Raised when a required parameter is not provided.

    For example:

    >>> raise RequirementError(name='givenname')
    Traceback (most recent call last):
      ...
    RequirementError: 'givenname' is required
    """

    errno = 3007
    format = _("'%(name)s' is required")


class ConversionError(InvocationError):
    """
    **3008** Raised when parameter value can't be converted to correct type.

    For example:

    >>> raise ConversionError(name='age', error=_(u'must be an integer'))
    Traceback (most recent call last):
      ...
    ConversionError: invalid 'age': must be an integer
    """

    errno = 3008
    format = _("invalid '%(name)s': %(error)s")


class ValidationError(InvocationError):
    """
    **3009** Raised when a parameter value fails a validation rule.

    For example:

    >>> raise ValidationError(name='sn', error=_(u'can be at most 128 characters'))
    Traceback (most recent call last):
      ...
    ValidationError: invalid 'sn': can be at most 128 characters
    """

    errno = 3009
    format = _("invalid '%(name)s': %(error)s")


class NoSuchNamespaceError(InvocationError):
    """
    **3010** Raised when an unknown namespace is requested.

    For example:

    >>> raise NoSuchNamespaceError(name='Plugins')
    Traceback (most recent call last):
      ...
    NoSuchNamespaceError: api has no such namespace: 'Plugins'
    """

    errno = 3010
    format = _("api has no such namespace: '%(name)s'")


class PasswordMismatch(InvocationError):
    """
    **3011** Raise when password and password confirmation don't match.
    """

    errno = 3011
    format = _('Passwords do not match')


class NotImplementedError(InvocationError):
    """
    **3012** Raise when a function hasn't been implemented.
    """

    errno = 3012
    format = _('Command not implemented')


class NotConfiguredError(InvocationError):
    """
    **3013** Raise when there is no configuration
    """

    errno = 3013
    format = _('Client is not configured. Run ipa-client-install.')


class PromptFailed(InvocationError):
    """
    **3014** Raise when an interactive prompt failed.
    """

    errno = 3014
    format = _('Could not get %(name)s interactively')


class DeprecationError(InvocationError):
    """
    **3015** Raise when a command has been deprecated

    For example:

    >>> raise DeprecationError(name='hbacrule_add_sourcehost')
    Traceback (most recent call last):
      ...
    DeprecationError: Command 'hbacrule_add_sourcehost' has been deprecated
    """
    errno = 3015
    format = _("Command '%(name)s' has been deprecated")

class NotAForestRootError(InvocationError):
    """
    **3016** Raised when an attempt to establish trust is done against non-root domain
             Forest root domain has the same name as the forest itself

    For example:

    >>> raise NotAForestRootError(forest='example.test', domain='jointops.test')
    Traceback (most recent call last):
      ...
    NotAForestRootError: Domain 'jointops.test' is not a root domain for forest 'example.test'
    """

    errno = 3016
    format = _("Domain '%(domain)s' is not a root domain for forest '%(forest)s'")

##############################################################################
# 4000 - 4999: Execution errors

class ExecutionError(PublicError):
    """
    **4000** Base class for execution errors (*4000 - 4999*).
    """

    errno = 4000

class NotFound(ExecutionError):
    """
    **4001** Raised when an entry is not found.

    For example:

    >>> raise NotFound(reason='no such user')
    Traceback (most recent call last):
      ...
    NotFound: no such user

    """

    errno = 4001
    rval = 2
    format = _('%(reason)s')

class DuplicateEntry(ExecutionError):
    """
    **4002** Raised when an entry already exists.

    For example:

    >>> raise DuplicateEntry
    Traceback (most recent call last):
      ...
    DuplicateEntry: This entry already exists

    """

    errno = 4002
    format = _('This entry already exists')

class HostService(ExecutionError):
    """
    **4003** Raised when a host service principal is requested

    For example:

    >>> raise HostService
    Traceback (most recent call last):
      ...
    HostService: You must enroll a host in order to create a host service

    """

    errno = 4003
    format = _('You must enroll a host in order to create a host service')

class MalformedServicePrincipal(ExecutionError):
    """
    **4004** Raised when a service principal is not of the form: service/fully-qualified host name

    For example:

    >>> raise MalformedServicePrincipal(reason=_('missing service'))
    Traceback (most recent call last):
      ...
    MalformedServicePrincipal: Service principal is not of the form: service/fully-qualified host name: missing service

    """

    errno = 4004
    format = _('Service principal is not of the form: service/fully-qualified host name: %(reason)s')

class RealmMismatch(ExecutionError):
    """
    **4005** Raised when the requested realm does not match the IPA realm

    For example:

    >>> raise RealmMismatch
    Traceback (most recent call last):
      ...
    RealmMismatch: The realm for the principal does not match the realm for this IPA server

    """

    errno = 4005
    format = _('The realm for the principal does not match the realm for this IPA server')

class RequiresRoot(ExecutionError):
    """
    **4006** Raised when a command requires the unix super-user to run

    For example:

    >>> raise RequiresRoot
    Traceback (most recent call last):
      ...
    RequiresRoot: This command requires root access

    """

    errno = 4006
    format = _('This command requires root access')

class AlreadyPosixGroup(ExecutionError):
    """
    **4007** Raised when a group is already a posix group

    For example:

    >>> raise AlreadyPosixGroup
    Traceback (most recent call last):
      ...
    AlreadyPosixGroup: This is already a posix group

    """

    errno = 4007
    format = _('This is already a posix group')

class MalformedUserPrincipal(ExecutionError):
    """
    **4008** Raised when a user principal is not of the form: user@REALM

    For example:

    >>> raise MalformedUserPrincipal(principal='jsmith@@EXAMPLE.COM')
    Traceback (most recent call last):
      ...
    MalformedUserPrincipal: Principal is not of the form user@REALM: 'jsmith@@EXAMPLE.COM'

    """

    errno = 4008
    format = _("Principal is not of the form user@REALM: '%(principal)s'")

class AlreadyActive(ExecutionError):
    """
    **4009** Raised when an entry is made active that is already active

    For example:

    >>> raise AlreadyActive()
    Traceback (most recent call last):
      ...
    AlreadyActive: This entry is already enabled

    """

    errno = 4009
    format = _('This entry is already enabled')

class AlreadyInactive(ExecutionError):
    """
    **4010** Raised when an entry is made inactive that is already inactive

    For example:

    >>> raise AlreadyInactive()
    Traceback (most recent call last):
      ...
    AlreadyInactive: This entry is already disabled

    """

    errno = 4010
    format = _('This entry is already disabled')

class HasNSAccountLock(ExecutionError):
    """
    **4011** Raised when an entry has the nsAccountLock attribute set

    For example:

    >>> raise HasNSAccountLock()
    Traceback (most recent call last):
      ...
    HasNSAccountLock: This entry cannot be enabled or disabled

    """

    errno = 4011
    format = _('This entry cannot be enabled or disabled')

class NotGroupMember(ExecutionError):
    """
    **4012** Raised when a non-member is attempted to be removed from a group

    For example:

    >>> raise NotGroupMember()
    Traceback (most recent call last):
      ...
    NotGroupMember: This entry is not a member

    """

    errno = 4012
    format = _('This entry is not a member')

class RecursiveGroup(ExecutionError):
    """
    **4013** Raised when a group is added as a member of itself

    For example:

    >>> raise RecursiveGroup()
    Traceback (most recent call last):
      ...
    RecursiveGroup: A group may not be a member of itself

    """

    errno = 4013
    format = _('A group may not be a member of itself')

class AlreadyGroupMember(ExecutionError):
    """
    **4014** Raised when a member is attempted to be re-added to a group

    For example:

    >>> raise AlreadyGroupMember()
    Traceback (most recent call last):
      ...
    AlreadyGroupMember: This entry is already a member

    """

    errno = 4014
    format = _('This entry is already a member')

class Base64DecodeError(ExecutionError):
    """
    **4015** Raised when a base64-encoded blob cannot decoded

    For example:

    >>> raise Base64DecodeError(reason=_('Incorrect padding'))
    Traceback (most recent call last):
      ...
    Base64DecodeError: Base64 decoding failed: Incorrect padding

    """

    errno = 4015
    format = _('Base64 decoding failed: %(reason)s')

class RemoteRetrieveError(ExecutionError):
    """
    **4016** Raised when retrieving data from a remote server fails

    For example:

    >>> raise RemoteRetrieveError(reason=_("Failed to get certificate chain."))
    Traceback (most recent call last):
      ...
    RemoteRetrieveError: Failed to get certificate chain.

    """

    errno = 4016
    format = _('%(reason)s')

class SameGroupError(ExecutionError):
    """
    **4017** Raised when adding a group as a member of itself

    For example:

    >>> raise SameGroupError()
    Traceback (most recent call last):
      ...
    SameGroupError: A group may not be added as a member of itself

    """

    errno = 4017
    format = _('A group may not be added as a member of itself')

class DefaultGroupError(ExecutionError):
    """
    **4018** Raised when removing the default user group

    For example:

    >>> raise DefaultGroupError()
    Traceback (most recent call last):
      ...
    DefaultGroupError: The default users group cannot be removed

    """

    errno = 4018
    format = _('The default users group cannot be removed')


class ManagedGroupError(ExecutionError):
    """
    **4020** Raised when a managed group is deleted

    For example:

    >>> raise ManagedGroupError()
    Traceback (most recent call last):
      ...
    ManagedGroupError: Deleting a managed group is not allowed. It must be detached first.
    """

    errno = 4020
    format = _('Deleting a managed group is not allowed. It must be detached first.')

class ManagedPolicyError(ExecutionError):
    """
    **4021** Raised when password policy is assigned to a managed group

    For example:

    >>> raise ManagedPolicyError()
    Traceback (most recent call last):
      ...
    ManagedPolicyError: A managed group cannot have a password policy.
    """

    errno = 4021
    format = _('A managed group cannot have a password policy.')


class FileError(ExecutionError):
    """
    **4022** Errors when dealing with files

    For example:

    >>> raise FileError(reason=_("cannot write file \'test\'"))
    Traceback (most recent call last):
      ...
    FileError: cannot write file 'test'
    """

    errno = 4022
    format = _('%(reason)s')


class NoCertificateError(ExecutionError):
    """
    **4023** Raised when trying to retrieve a certificate that doesn't exist.

    For example:

    >>> raise NoCertificateError(entry='ipa.example.com')
    Traceback (most recent call last):
      ...
    NoCertificateError: 'ipa.example.com' doesn't have a certificate.
    """

    errno = 4023
    format = _('\'%(entry)s\' doesn\'t have a certificate.')


class ManagedGroupExistsError(ExecutionError):
    """
    **4024** Raised when adding a user and its managed group exists

    For example:

    >>> raise ManagedGroupExistsError(group=u'engineering')
    Traceback (most recent call last):
      ...
    ManagedGroupExistsError: Unable to create private group. A group 'engineering' already exists.
    """

    errno = 4024
    format = _('Unable to create private group. A group \'%(group)s\' already exists.')


class ReverseMemberError(ExecutionError):
    """
    **4025** Raised when verifying that all reverse members have been added or removed.

    For example:

    >>> raise ReverseMemberError(verb=_('added'), exc=_("Group 'foo' not found."))
    Traceback (most recent call last):
      ...
    ReverseMemberError: A problem was encountered when verifying that all members were added: Group 'foo' not found.
    """

    errno = 4025
    format = _('A problem was encountered when verifying that all members were %(verb)s: %(exc)s')


class AttrValueNotFound(ExecutionError):
    """
    **4026** Raised when an Attribute/Value pair is not found.

    For example:

    >>> raise AttrValueNotFound(attr='ipasudoopt', value='authenticate')
    Traceback (most recent call last):
      ...
    AttrValueNotFound: ipasudoopt does not contain 'authenticate'

    """

    errno = 4026
    rval = 1
    format = _('%(attr)s does not contain \'%(value)s\'')


class SingleMatchExpected(ExecutionError):
    """
    **4027** Raised when a search should return a single match

    For example:

    >>> raise SingleMatchExpected(found=9)
    Traceback (most recent call last):
      ...
    SingleMatchExpected: The search criteria was not specific enough. Expected 1 and found 9.
    """

    errno = 4027
    rval = 1
    format = _('The search criteria was not specific enough. Expected 1 and found %(found)d.')


class AlreadyExternalGroup(ExecutionError):
    """
    **4028** Raised when a group is already an external member group

    For example:

    >>> raise AlreadyExternalGroup
    Traceback (most recent call last):
      ...
    AlreadyExternalGroup: This group already allows external members

    """

    errno = 4028
    format = _('This group already allows external members')

class ExternalGroupViolation(ExecutionError):
    """
    **4029** Raised when a group is already an external member group
             and an attempt is made to use it as posix group

    For example:

    >>> raise ExternalGroupViolation
    Traceback (most recent call last):
      ...
    ExternalGroupViolation: This group cannot be posix because it is external

    """

    errno = 4029
    format = _('This group cannot be posix because it is external')

class PosixGroupViolation(ExecutionError):
    """
    **4030** Raised when a group is already a posix group
             and cannot be converted to external

    For example:

    >>> raise PosixGroupViolation
    Traceback (most recent call last):
      ...
    PosixGroupViolation: This is already a posix group and cannot be converted to external one

    """

    errno = 4030
    format = _('This is already a posix group and cannot be converted to external one')

class EmptyResult(NotFound):
    """
    **4031** Raised when a LDAP search returned no results.

    For example:

    >>> raise EmptyResult(reason='no matching entry found')
    Traceback (most recent call last):
      ...
    EmptyResult: no matching entry found

    """

    errno = 4031

class InvalidDomainLevelError(ExecutionError):
    """
    **4032** Raised when a operation could not be completed due to a invalid
             domain level.

    For example:

    >>> raise InvalidDomainLevelError(reason='feature requires domain level 4')
    Traceback (most recent call last):
      ...
    InvalidDomainLevelError: feature requires domain level 4

    """

    errno = 4032
    format = _('%(reason)s')


class ServerRemovalError(ExecutionError):
    """
    **4033** Raised when a removal of IPA server from managed topology fails

    For example:

    >>> raise ServerRemovalError(reason='Removal disconnects topology')
    Traceback (most recent call last):
      ...
    ServerRemovalError: Server removal aborted: Removal disconnects topology

    """

    errno = 4033
    format = _('Server removal aborted: %(reason)s.')


class OperationNotSupportedForPrincipalType(ExecutionError):
    """
    **4034** Raised when an operation is not supported for a principal type
    """

    errno = 4034
    format = _(
        '%(operation)s is not supported for %(principal_type)s principals')


class HTTPRequestError(RemoteRetrieveError):
    """
    **4035** Raised when an HTTP request fails.  Includes the response
    status in the ``status`` attribute.
    """

    errno = 4035
    format = _('Request failed with status %(status)s: %(reason)s')


class BuiltinError(ExecutionError):
    """
    **4100** Base class for builtin execution errors (*4100 - 4199*).
    """

    errno = 4100


class HelpError(BuiltinError):
    """
    **4101** Raised when requesting help for an unknown topic.

    For example:

    >>> raise HelpError(topic='newfeature')
    Traceback (most recent call last):
      ...
    HelpError: no command nor help topic 'newfeature'
    """

    errno = 4101
    format = _("no command nor help topic '%(topic)s'")


class LDAPError(ExecutionError):
    """
    **4200** Base class for LDAP execution errors (*4200 - 4299*).
    """

    errno = 4200


class MidairCollision(ExecutionError):
    """
    **4201** Raised when a change collides with another change

    For example:

    >>> raise MidairCollision()
    Traceback (most recent call last):
      ...
    MidairCollision: change collided with another change
    """

    errno = 4201
    format = _('change collided with another change')


class EmptyModlist(ExecutionError):
    """
    **4202** Raised when an LDAP update makes no changes

    For example:

    >>> raise EmptyModlist()
    Traceback (most recent call last):
      ...
    EmptyModlist: no modifications to be performed
    """

    errno = 4202
    format = _('no modifications to be performed')


class DatabaseError(ExecutionError):
    """
    **4203** Raised when an LDAP error is not otherwise handled

    For example:

    >>> raise DatabaseError(desc=_("Can't contact LDAP server"), info=_('Info goes here'))
    Traceback (most recent call last):
      ...
    DatabaseError: Can't contact LDAP server: Info goes here
    """

    errno = 4203
    format = _('%(desc)s: %(info)s')


class LimitsExceeded(ExecutionError):
    """
    **4204** Raised when search limits are exceeded.

    For example:

    >>> raise LimitsExceeded()
    Traceback (most recent call last):
      ...
    LimitsExceeded: limits exceeded for this query
    """

    errno = 4204
    format = _('limits exceeded for this query')

class ObjectclassViolation(ExecutionError):
    """
    **4205** Raised when an entry is missing a required attribute or objectclass

    For example:

    >>> raise ObjectclassViolation(info=_('attribute "krbPrincipalName" not allowed'))
    Traceback (most recent call last):
      ...
    ObjectclassViolation: attribute "krbPrincipalName" not allowed
    """

    errno = 4205
    format = _('%(info)s')

class NotAllowedOnRDN(ExecutionError):
    """
    **4206** Raised when an RDN value is modified.

    For example:

    >>> raise NotAllowedOnRDN()
    Traceback (most recent call last):
      ...
    NotAllowedOnRDN: modifying primary key is not allowed
    """

    errno = 4206
    format = _('modifying primary key is not allowed')


class OnlyOneValueAllowed(ExecutionError):
    """
    **4207** Raised when trying to set more than one value to single-value attributes

    For example:

    >> raise OnlyOneValueAllowed(attr='ipasearchtimelimit')
    Traceback (most recent call last):
      ...
    OnlyOneValueAllowed: ipasearchtimelimit: Only one value allowed.
    """

    errno = 4207
    format = _('%(attr)s: Only one value allowed.')


class InvalidSyntax(ExecutionError):
    """
    **4208** Raised when an value does not match the required syntax

    For example:

    >> raise InvalidSyntax(attr='ipahomesrootdir')
    Traceback (most recent call last):
      ...
    InvalidSyntax: ipahomesrootdir: Invalid syntax
    """

    errno = 4208
    format = _('%(attr)s: Invalid syntax.')


class BadSearchFilter(ExecutionError):
    """
    **4209** Raised when an invalid LDAP search filter is used

    For example:

    >>> raise BadSearchFilter(info=_('invalid syntax'))
    Traceback (most recent call last):
      ...
    BadSearchFilter: Bad search filter invalid syntax
    """

    errno = 4209
    format = _('Bad search filter %(info)s')


class NotAllowedOnNonLeaf(ExecutionError):
    """
    **4210** Raised when operation is not allowed on a non-leaf entry

    For example:

    >>> raise NotAllowedOnNonLeaf()
    Traceback (most recent call last):
      ...
    NotAllowedOnNonLeaf: Not allowed on non-leaf entry
    """

    errno = 4210
    format = _('Not allowed on non-leaf entry')


class DatabaseTimeout(DatabaseError):
    """
    **4211** Raised when an LDAP call times out

    For example:

    >>> raise DatabaseTimeout()
    Traceback (most recent call last):
      ...
    DatabaseTimeout: LDAP timeout
    """

    errno = 4211
    format = _('LDAP timeout')


class TaskTimeout(DatabaseError):
    """
    **4213** Raised when an LDAP task times out

    For example:

    >>> raise TaskTimeout(task='Automember', task_dn='')
    Traceback (most recent call last):
      ...
    TaskTimeout: Automember LDAP task timeout, Task DN: ''
    """

    errno = 4213
    format = _("%(task)s LDAP task timeout, Task DN: '%(task_dn)s'")


class TimeLimitExceeded(LimitsExceeded):
    """
    **4214** Raised when time limit for the operation is exceeded.
    """

    errno = 4214
    format = _('Configured time limit exceeded')


class SizeLimitExceeded(LimitsExceeded):
    """
    **4215** Raised when size limit for the operation is exceeded.
    """

    errno = 4215
    format = _('Configured size limit exceeded')


class AdminLimitExceeded(LimitsExceeded):
    """
    **4216** Raised when server limit imposed by administrative authority was
    exceeded
    """

    errno = 4216
    format = _('Configured administrative server limit exceeded')


class CertificateError(ExecutionError):
    """
    **4300** Base class for Certificate execution errors (*4300 - 4399*).
    """

    errno = 4300


class CertificateOperationError(CertificateError):
    """
    **4301** Raised when a certificate operation cannot be completed

    For example:

    >>> raise CertificateOperationError(error=_(u'bad serial number'))
    Traceback (most recent call last):
      ...
    CertificateOperationError: Certificate operation cannot be completed: bad serial number

    """

    errno = 4301
    format = _('Certificate operation cannot be completed: %(error)s')

class CertificateFormatError(CertificateError):
    """
    **4302** Raised when a certificate is badly formatted

    For example:

    >>> raise CertificateFormatError(error=_(u'improperly formated DER-encoded certificate'))
    Traceback (most recent call last):
      ...
    CertificateFormatError: Certificate format error: improperly formated DER-encoded certificate

    """

    errno = 4302
    format = _('Certificate format error: %(error)s')


class MutuallyExclusiveError(ExecutionError):
    """
    **4303** Raised when an operation would result in setting two attributes which are mutually exlusive.

    For example:

    >>> raise MutuallyExclusiveError(reason=_(u'hosts may not be added when hostcategory=all'))
    Traceback (most recent call last):
      ...
    MutuallyExclusiveError: hosts may not be added when hostcategory=all

    """

    errno = 4303
    format = _('%(reason)s')


class NonFatalError(ExecutionError):
    """
    **4304** Raised when part of an operation succeeds and the part that failed isn't critical.

    For example:

    >>> raise NonFatalError(reason=_(u'The host was added but the DNS update failed'))
    Traceback (most recent call last):
      ...
    NonFatalError: The host was added but the DNS update failed

    """

    errno = 4304
    format = _('%(reason)s')


class AlreadyRegisteredError(ExecutionError):
    """
    **4305** Raised when registering a user that is already registered.

    For example:

    >>> raise AlreadyRegisteredError()
    Traceback (most recent call last):
      ...
    AlreadyRegisteredError: Already registered

    """

    errno = 4305
    format = _('Already registered')


class NotRegisteredError(ExecutionError):
    """
    **4306** Raised when not registered and a registration is required

    For example:
    >>> raise NotRegisteredError()
    Traceback (most recent call last):
      ...
    NotRegisteredError: Not registered yet

    """

    errno = 4306
    format = _('Not registered yet')


class DependentEntry(ExecutionError):
    """
    **4307** Raised when an entry being deleted has dependencies

    For example:
    >>> raise DependentEntry(label=u'SELinux User Map', key=u'test', dependent=u'test1')
    Traceback (most recent call last):
      ...
    DependentEntry: test cannot be deleted because SELinux User Map test1 requires it

    """

    errno = 4307
    format = _('%(key)s cannot be deleted because %(label)s %(dependent)s requires it')


class LastMemberError(ExecutionError):
    """
    **4308** Raised when an entry being deleted or disabled is last member of a protected group

    For example:
    >>> raise LastMemberError(key=u'admin', label=u'group', container=u'admins')
    Traceback (most recent call last):
      ...
    LastMemberError: admin cannot be deleted or disabled because it is the last member of group admins

    """

    errno = 4308
    format = _('%(key)s cannot be deleted or disabled because it is the last member of %(label)s %(container)s')


class ProtectedEntryError(ExecutionError):
    """
    **4309** Raised when an entry being deleted or modified in a forbidden way is protected

    For example:
    >>> raise ProtectedEntryError(label=u'group', key=u'admins', reason=_(u'privileged group'))
    Traceback (most recent call last):
      ...
    ProtectedEntryError: group admins cannot be deleted/modified: privileged group

    """

    errno = 4309
    format = _('%(label)s %(key)s cannot be deleted/modified: %(reason)s')


class CertificateInvalidError(CertificateError):
    """
    **4310** Raised when a certificate is not valid

    For example:

    >>> raise CertificateInvalidError(name=_(u'CA'))
    Traceback (most recent call last):
      ...
    CertificateInvalidError: CA certificate is not valid

    """

    errno = 4310
    format = _('%(name)s certificate is not valid')


class SchemaUpToDate(ExecutionError):
    """
    **4311** Raised by server when client asks for metadata but
    already has current version. Exception's attribute 'fingerprint'
    identitfies schema version to use. Attribute 'ttl' specifies how
    long (in seconds) before client should check for schema update.

    For example:
    >>> raise SchemaUpToDate(fingerprint=u'deadbeef', ttl=3600)
    Traceback (most recent call last):
      ...
    SchemaUpToDate: Schema is up to date (FP 'deadbeef', TTL 3600 s)
    """

    errno = 4311
    format = _("Schema is up to date (FP '%(fingerprint)s', TTL %(ttl)s s)")


class DNSError(ExecutionError):
    """
    **4400** Base class for DNS execution errors (*4400 - 4499*).
    These are typically wrapper exceptions around dns.exception.DNSException.
    """

    errno = 4400


class DNSNotARecordError(DNSError):
    """
    **4019** Raised when a hostname is not a DNS A/AAAA record

    For example:

    >>> raise DNSNotARecordError(hostname='x')
    Traceback (most recent call last):
      ...
    DNSNotARecordError: Host 'x' does not have corresponding DNS A/AAAA record

    """

    errno = 4019  # this exception was defined before DNSError
    format = _(
        'Host \'%(hostname)s\' does not have corresponding DNS A/AAAA record')


class DNSDataMismatch(DNSError):
    """
    **4212** Raised when an DNS query didn't return expected answer
    in a configured time limit.

    For example:

    >>> raise DNSDataMismatch(expected="zone3.test. 86400 IN A 192.0.2.1", \
                              got="zone3.test. 86400 IN A 192.168.1.1")
    Traceback (most recent call last):
      ...
    DNSDataMismatch: DNS check failed: Expected {zone3.test. 86400 IN A 192.0.2.1} got {zone3.test. 86400 IN A 192.168.1.1}
    """

    errno = 4212  # this exception was defined before DNSError
    format = _('DNS check failed: Expected {%(expected)s} got {%(got)s}')


class DNSResolverError(DNSError):
    """
    **4401** Wrapper around dns.exception.DNSException.
    Raised when an error occured in dns.resolver.

    For example:
    >>> raise DNSResolverError(exception=ValueError("this is bad"))
    Traceback (most recent call last):
      ...
    DNSResolverError: this is bad
    """

    errno = 4401
    format = _('%(exception)s')

class TrustError(ExecutionError):
    """
    **4500** Base class for trust execution errors (*4500 - 4599*).
    These are typically instantiated when there is an error in establishing or
    modifying a trust to another forest.
    """

    errno = 4500

class TrustTopologyConflictError(TrustError):
    """
    **4501** Raised when an attempt to establish trust fails with a topology
             conflict against another forest the target forest trusts

    For example:

    >>> raise TrustTopologyConflictError(forest='example.test',
                                         conflict='my.ad.test',
                                         domains=['ad.test'])
    Traceback (most recent call last):
      ...
    TrustTopologyConflictError: Forest 'example.test' has existing trust to forest(s) ['ad.test'] which prevents a trust to 'my.ad.test'
    """

    errno = 4501
    format = _("Forest '%(forest)s' has existing trust to forest(s) "
               "%(domains)s which prevents a trust to '%(conflict)s'")


##############################################################################
# 5000 - 5999: Generic errors

class GenericError(PublicError):
    """
    **5000** Base class for errors that don't fit elsewhere (*5000 - 5999*).
    """

    errno = 5000



public_errors = tuple(sorted(
    messages.iter_messages(globals(), PublicError), key=lambda E: E.errno))

if __name__ == '__main__':
    messages.print_report('public errors', public_errors)



#
# Copyright (C) 2016  FreeIPA Contributors see COPYING for license
#

import re
from ipalib import LocalOrRemote, _, ngettext
from ipalib.output import Output, summary
from ipalib import Flag
from ipalib.plugable import Registry

register = Registry()

# FIXME: We should not let env return anything in_server
# when mode == 'production'.  This would allow an attacker to see the
# configuration of the server, potentially revealing compromising
# information.  However, it's damn handy for testing/debugging.


class env(LocalOrRemote):
    __doc__ = _('Show environment variables.')

    msg_summary = _('%(count)d variables')

    takes_args = (
        'variables*',
    )

    takes_options = LocalOrRemote.takes_options + (
        Flag(
            'all',
            cli_name='all',
            doc=_('retrieve and print all attributes from the server. '
                  'Affects command output.'),
            exclude='webui',
            flags=['no_option', 'no_output'],
            default=True,
        ),
    )

    has_output = (
        Output(
            'result',
            type=dict,
            doc=_('Dictionary mapping variable name to value'),
        ),
        Output(
            'total',
            type=int,
            doc=_('Total number of variables env (>= count)'),
            flags=['no_display'],
        ),
        Output(
            'count',
            type=int,
            doc=_('Number of variables returned (<= total)'),
            flags=['no_display'],
        ),
        summary,
    )

    def __find_keys(self, variables):
        keys = set()
        for query in variables:
            if '*' in query:
                pat = re.compile(query.replace('*', '.*') + '$')
                for key in self.env:
                    if pat.match(key):
                        keys.add(key)
            elif query in self.env:
                keys.add(query)
        return keys

    def execute(self, variables=None, **options):
        if variables is None:
            keys = self.env
        else:
            keys = self.__find_keys(variables)
        ret = dict(
            result=dict(
                (key, self.env[key]) for key in keys
            ),
            count=len(keys),
            total=len(self.env),
        )
        if len(keys) > 1:
            ret['summary'] = self.msg_summary % ret
        else:
            ret['summary'] = None
        return ret


class plugins(LocalOrRemote):
    __doc__ = _('Show all loaded plugins.')

    msg_summary = ngettext(
        '%(count)d plugin loaded', '%(count)d plugins loaded', 0
    )

    takes_options = LocalOrRemote.takes_options + (
        Flag(
            'all',
            cli_name='all',
            doc=_('retrieve and print all attributes from the server. '
                  'Affects command output.'),
            exclude='webui',
            flags=['no_option', 'no_output'],
            default=True,
        ),
    )

    has_output = (
        Output('result', dict, 'Dictionary mapping plugin names to bases'),
        Output(
            'count',
            type=int,
            doc=_('Number of plugins loaded'),
        ),
        summary,
    )

    def execute(self, **options):
        result = {}
        for namespace in self.api:
            for plugin in self.api[namespace]():
                cls = type(plugin)
                key = '{}.{}'.format(cls.__module__, cls.__name__)
                result.setdefault(key, []).append(namespace.decode('utf-8'))

        return dict(
            result=result,
            count=len(result),
        )



from __future__ import print_function
import sys

print(
    "ipalib.pkcs10 module is deprecated and will be removed in FreeIPA 4.6. "
    "To load CSRs, please, use python-cryptography instead.",
    file=sys.stderr
)



#
# Copyright (C) 2016  FreeIPA Contributors see COPYING for license
#

"""
Host name installer module
"""

from ipapython.install import typing
from ipapython.install.core import knob
from ipapython.ipautil import CheckedIPAddress

from . import service
from .service import prepare_only


class HostNameInstallInterface(service.ServiceInstallInterface):
    """
    Interface common to all service installers which create DNS address
    records for `host_name`
    """

    ip_addresses = knob(
        # pylint: disable=invalid-sequence-index
        typing.List[CheckedIPAddress], None,
        description="Specify IP address that should be added to DNS. This "
                    "option can be used multiple times",
        cli_names='--ip-address',
        cli_metavar='IP_ADDRESS',
    )
    ip_addresses = prepare_only(ip_addresses)

    @ip_addresses.validator
    def ip_addresses(self, values):
        for value in values:
            try:
                CheckedIPAddress(value)
            except Exception as e:
                raise ValueError("invalid IP address {0}: {1}".format(
                    value, e))

    all_ip_addresses = knob(
        None,
        description="All routable IP addresses configured on any interface "
                    "will be added to DNS",
    )
    all_ip_addresses = prepare_only(all_ip_addresses)

    no_host_dns = knob(
        None,
        description="Do not use DNS for hostname lookup during installation",
    )
    no_host_dns = prepare_only(no_host_dns)

    no_wait_for_dns = knob(
        None,
        description="do not wait until the host is resolvable in DNS",
    )
    no_wait_for_dns = prepare_only(no_wait_for_dns)



# Authors: Rob Crittenden <rcritten@redhat.com>
#          David Kupka <dkupka@redhat.com>
#
# Copyright (C) 2010  Red Hat
# see file 'COPYING' for use and warranty information
#
# This program is free software; you can redistribute it and/or modify
# it under the terms of the GNU General Public License as published by
# the Free Software Foundation, either version 3 of the License, or
# (at your option) any later version.
#
# This program is distributed in the hope that it will be useful,
# but WITHOUT ANY WARRANTY; without even the implied warranty of
# MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
# GNU General Public License for more details.
#
# You should have received a copy of the GNU General Public License
# along with this program.  If not, see <http://www.gnu.org/licenses/>.
#

# Some certmonger functions, mostly around updating the request file.
# This is used so we can add tracking to the Apache and 389-ds
# server certificates created during the IPA server installation.

from __future__ import print_function, absolute_import

import logging
import os
import time
import dbus
import shlex
import subprocess
import tempfile
from ipalib import api
from ipalib.constants import CA_DBUS_TIMEOUT
from ipapython.dn import DN
from ipaplatform.paths import paths
from ipaplatform import services

logger = logging.getLogger(__name__)

DBUS_CM_PATH = '/org/fedorahosted/certmonger'
DBUS_CM_IF = 'org.fedorahosted.certmonger'
DBUS_CM_NAME = 'org.fedorahosted.certmonger'
DBUS_CM_REQUEST_IF = 'org.fedorahosted.certmonger.request'
DBUS_CM_CA_IF = 'org.fedorahosted.certmonger.ca'
DBUS_PROPERTY_IF = 'org.freedesktop.DBus.Properties'


class _cm_dbus_object(object):
    """
    Auxiliary class for convenient DBus object handling.
    """
    def __init__(self, bus, parent, object_path, object_dbus_interface,
                 parent_dbus_interface=None, property_interface=False):
        """
        bus - DBus bus object, result of dbus.SystemBus() or dbus.SessionBus()
              Object is accesible over this DBus bus instance.
        object_path - path to requested object on DBus bus
        object_dbus_interface
        parent_dbus_interface
        property_interface - create DBus property interface? True or False
        """
        if bus is None or object_path is None or object_dbus_interface is None:
            raise RuntimeError(
                "bus, object_path and dbus_interface must not be None.")
        if parent_dbus_interface is None:
            parent_dbus_interface = object_dbus_interface
        self.bus = bus
        self.parent = parent
        self.path = object_path
        self.obj_dbus_if = object_dbus_interface
        self.parent_dbus_if = parent_dbus_interface
        self.obj = bus.get_object(parent_dbus_interface, object_path)
        self.obj_if = dbus.Interface(self.obj, object_dbus_interface)
        if property_interface:
            self.prop_if = dbus.Interface(self.obj, DBUS_PROPERTY_IF)


class _certmonger(_cm_dbus_object):
    """
    Create a connection to certmonger.
    By default use SystemBus. When not available use private connection
    over Unix socket.
    This solution is really ugly and should be removed as soon as DBus
    SystemBus is available at system install time.
    """
    timeout = 300

    def _start_private_conn(self):
        sock_filename = os.path.join(tempfile.mkdtemp(), 'certmonger')
        self._proc = subprocess.Popen([paths.CERTMONGER, '-n', '-L', '-P',
                                       sock_filename])
        for _t in range(0, self.timeout, 5):
            if os.path.exists(sock_filename):
                return "unix:path=%s" % sock_filename
            time.sleep(5)
        self._stop_private_conn()
        raise RuntimeError("Failed to start certmonger: Timed out")

    def _stop_private_conn(self):
        if self._proc:
            retcode = self._proc.poll()
            if retcode is not None:
                return
            self._proc.terminate()
            for _t in range(0, self.timeout, 5):
                retcode = self._proc.poll()
                if retcode is not None:
                    return
                time.sleep(5)
            logger.error("Failed to stop certmonger.")

    def __del__(self):
        self._stop_private_conn()

    def __init__(self):
        self._proc = None
        self._bus = None
        try:
            self._bus = dbus.SystemBus()
        except dbus.DBusException as e:
            err_name = e.get_dbus_name()
            if err_name not in ['org.freedesktop.DBus.Error.NoServer',
                                'org.freedesktop.DBus.Error.FileNotFound']:
                logger.error("Failed to connect to certmonger over "
                             "SystemBus: %s", e)
                raise
            try:
                self._private_sock = self._start_private_conn()
                self._bus = dbus.connection.Connection(self._private_sock)
            except dbus.DBusException as e:
                logger.error("Failed to connect to certmonger over "
                             "private socket: %s", e)
                raise
        else:
            try:
                self._bus.get_name_owner(DBUS_CM_NAME)
            except dbus.DBusException:
                try:
                    services.knownservices.certmonger.start()
                except Exception as e:
                    logger.error("Failed to start certmonger: %s", e)
                    raise

                for _t in range(0, self.timeout, 5):
                    try:
                        self._bus.get_name_owner(DBUS_CM_NAME)
                        break
                    except dbus.DBusException:
                        pass
                    time.sleep(5)
                    raise RuntimeError('Failed to start certmonger')

        super(_certmonger, self).__init__(self._bus, None, DBUS_CM_PATH,
                                          DBUS_CM_IF)


def _get_requests(criteria=dict()):
    """
    Get all requests that matches the provided criteria.
    """
    if not isinstance(criteria, dict):
        raise TypeError('"criteria" must be dict.')

    cm = _certmonger()
    requests = []
    requests_paths = []
    if 'nickname' in criteria:
        request_path = cm.obj_if.find_request_by_nickname(criteria['nickname'])
        if request_path:
            requests_paths = [request_path]
    else:
        requests_paths = cm.obj_if.get_requests()

    for request_path in requests_paths:
        request = _cm_dbus_object(cm.bus, cm, request_path, DBUS_CM_REQUEST_IF,
                                  DBUS_CM_IF, True)
        for criterion in criteria:
            if criterion == 'ca-name':
                ca_path = request.obj_if.get_ca()
                if ca_path is None:
                    raise RuntimeError("certmonger CA '%s' is not defined" %
                                       criteria.get('ca-name'))
                ca = _cm_dbus_object(cm.bus, cm, ca_path, DBUS_CM_CA_IF,
                                     DBUS_CM_IF)
                value = ca.obj_if.get_nickname()
            else:
                value = request.prop_if.Get(DBUS_CM_REQUEST_IF, criterion)
            if value != criteria[criterion]:
                break
        else:
            requests.append(request)

    return requests


def _get_request(criteria):
    """
    Find request that matches criteria.
    If 'nickname' is specified other criteria are ignored because 'nickname'
    uniquely identify single request.
    When multiple or none request matches specified criteria RuntimeError is
    raised.
    """
    requests = _get_requests(criteria)
    if len(requests) == 0:
        return None
    elif len(requests) == 1:
        return requests[0]
    else:
        raise RuntimeError("Criteria expected to be met by 1 request, got %s."
                           % len(requests))


def get_request_value(request_id, directive):
    """
    Get property of request.
    """
    try:
        request = _get_request(dict(nickname=request_id))
    except RuntimeError as e:
        logger.error('Failed to get request: %s', e)
        raise
    if request:
        if directive == 'ca-name':
            ca_path = request.obj_if.get_ca()
            ca = _cm_dbus_object(request.bus, request, ca_path, DBUS_CM_CA_IF,
                                 DBUS_CM_IF)
            return ca.obj_if.get_nickname()
        else:
            return request.prop_if.Get(DBUS_CM_REQUEST_IF, directive)
    else:
        return None


def get_request_id(criteria):
    """
    If you don't know the certmonger request_id then try to find it by looking
    through all the requests.

    criteria is a tuple of key/value to search for. The more specific
    the better. An error is raised if multiple request_ids are returned for
    the same criteria.

    None is returned if none of the criteria match.
    """
    try:
        request = _get_request(criteria)
    except RuntimeError as e:
        logger.error('Failed to get request: %s', e)
        raise
    if request:
        return request.prop_if.Get(DBUS_CM_REQUEST_IF, 'nickname')
    else:
        return None


def get_requests_for_dir(dir):
    """
    Return a list containing the request ids for a given NSS database
    directory.
    """
    reqid = []
    criteria = {'cert-storage': 'NSSDB', 'key-storage': 'NSSDB',
                'cert-database': dir, 'key-database': dir, }
    requests = _get_requests(criteria)
    for request in requests:
        reqid.append(request.prop_if.Get(DBUS_CM_REQUEST_IF, 'nickname'))

    return reqid


def add_request_value(request_id, directive, value):
    """
    Add a new directive to a certmonger request file.
    """
    try:
        request = _get_request({'nickname': request_id})
    except RuntimeError as e:
        logger.error('Failed to get request: %s', e)
        raise
    if request:
        request.obj_if.modify({directive: value})


def add_principal(request_id, principal):
    """
    In order for a certmonger request to be renewable it needs a principal.

    When an existing certificate is added via start-tracking it won't have
    a principal.
    """
    add_request_value(request_id, 'template-principal', [principal])


def add_subject(request_id, subject):
    """
    In order for a certmonger request to be renwable it needs the subject
    set in the request file.

    When an existing certificate is added via start-tracking it won't have
    a subject_template set.
    """
    add_request_value(request_id, 'template-subject', subject)


def request_and_wait_for_cert(
        certpath, subject, principal, nickname=None, passwd_fname=None,
        dns=None, ca='IPA', profile=None,
        pre_command=None, post_command=None, storage='NSSDB', perms=None,
        resubmit_timeout=0):
    """Request certificate, wait and possibly resubmit failing requests

    Submit a cert request to certmonger and wait until the request has
    finished.

    With timeout, a failed request is resubmitted. During parallel replica
    installation, a request sometimes fails with CA_REJECTED or
    CA_UNREACHABLE. The error occurs when the master is either busy or some
    information haven't been replicated yet. Even a stuck request can be
    recovered, e.g. when permission and group information have been
    replicated.
    """
    req_id = request_cert(
        certpath, subject, principal, nickname, passwd_fname, dns, ca,
        profile, pre_command, post_command, storage, perms
    )

    deadline = time.time() + resubmit_timeout
    while True:  # until success, timeout, or error
        state = wait_for_request(req_id, api.env.replication_wait_timeout)
        ca_error = get_request_value(req_id, 'ca-error')
        if state == 'MONITORING' and ca_error is None:
            # we got a winner, exiting
            logger.debug("Cert request %s was successful", req_id)
            return req_id

        logger.debug(
            "Cert request %s failed: %s (%s)", req_id, state, ca_error
        )
        if state not in {'CA_REJECTED', 'CA_UNREACHABLE'}:
            # probably unrecoverable error
            logger.debug("Giving up on cert request %s", req_id)
            break
        if not resubmit_timeout:
            # no resubmit
            break
        if time.time() > deadline:
            logger.debug("Request %s reached resubmit dead line", req_id)
            break
        else:
            # sleep and resubmit
            logger.debug("Sleep and resubmit cert request %s", req_id)
            time.sleep(10)
            resubmit_request(req_id)

    raise RuntimeError(
        "Certificate issuance failed ({}: {})".format(state, ca_error)
    )


def request_cert(
        certpath, subject, principal, nickname=None, passwd_fname=None,
        dns=None, ca='IPA', profile=None,
        pre_command=None, post_command=None, storage='NSSDB', perms=None):
    """
    Execute certmonger to request a server certificate.

    ``dns``
        A sequence of DNS names to appear in SAN request extension.
    ``perms``
        A tuple of (cert, key) permissions in e.g., (0644,0660)
    """
    if storage == 'FILE':
        certfile, keyfile = certpath
        # This is a workaround for certmonger having different Subject
        # representation with NSS and OpenSSL
        # https://pagure.io/certmonger/issue/62
        subject = str(DN(*reversed(DN(subject))))
    else:
        certfile = certpath
        keyfile = certpath

    cm = _certmonger()
    ca_path = cm.obj_if.find_ca_by_nickname(ca)
    if not ca_path:
        raise RuntimeError('{} CA not found'.format(ca))
    request_parameters = dict(KEY_STORAGE=storage, CERT_STORAGE=storage,
                              CERT_LOCATION=certfile, KEY_LOCATION=keyfile,
                              SUBJECT=subject, CA=ca_path)
    if nickname:
        request_parameters["CERT_NICKNAME"] = nickname
        request_parameters["KEY_NICKNAME"] = nickname
    if principal:
        request_parameters['PRINCIPAL'] = [principal]
    if dns is not None and len(dns) > 0:
        request_parameters['DNS'] = dns
    if passwd_fname:
        request_parameters['KEY_PIN_FILE'] = passwd_fname
    if profile:
        request_parameters['ca-profile'] = profile

    certmonger_cmd_template = paths.CERTMONGER_COMMAND_TEMPLATE
    if pre_command:
        if not os.path.isabs(pre_command):
            pre_command = certmonger_cmd_template % (pre_command)
        request_parameters['cert-presave-command'] = pre_command
    if post_command:
        if not os.path.isabs(post_command):
            post_command = certmonger_cmd_template % (post_command)
        request_parameters['cert-postsave-command'] = post_command

    if perms:
        request_parameters['cert-perms'] = perms[0]
        request_parameters['key-perms'] = perms[1]

    result = cm.obj_if.add_request(request_parameters)
    try:
        if result[0]:
            request = _cm_dbus_object(cm.bus, cm, result[1], DBUS_CM_REQUEST_IF,
                                      DBUS_CM_IF, True)
        else:
            raise RuntimeError('add_request() returned False')
    except Exception as e:
        logger.error('Failed to create a new request: %s', e)
        raise
    return request.obj_if.get_nickname()


def start_tracking(
        certpath, ca='IPA', nickname=None, pin=None, pinfile=None,
        pre_command=None, post_command=None, profile=None, storage="NSSDB"):
    """
    Tell certmonger to track the given certificate in either a file or an NSS
    database. The certificate access can be protected by a password_file.

    This uses the generic certmonger command getcert so we can specify
    a different helper.

    :param certpath:
        The path to an NSS database or a tuple (PEM certificate, private key).
    :param ca:
        Nickanme of the CA for which the given certificate should be tracked.
    :param nickname:
        Nickname of the NSS certificate in ``certpath`` to be tracked.
    :param pin:
        The passphrase for either NSS database containing ``nickname`` or
        for the encrypted key in the ``certpath`` tuple.
    :param pinfile:
        Similar to ``pin`` parameter except this is a path to a file containing
        the required passphrase.
    :param pre_command:
        Specifies a command for certmonger to run before it renews a
        certificate. This command must reside in /usr/lib/ipa/certmonger
        to work with SELinux.
    :param post_command:
        Specifies a command for certmonger to run after it has renewed a
        certificate. This command must reside in /usr/lib/ipa/certmonger
        to work with SELinux.
    :param storage:
        One of "NSSDB" or "FILE", describes whether certmonger should use
        NSS or OpenSSL backend to track the certificate in ``certpath``
    :param profile:
        Which certificate profile should be used.
    :returns: certificate tracking nickname.
    """
    if storage == 'FILE':
        certfile, keyfile = certpath
    else:
        certfile = certpath
        keyfile = certpath

    cm = _certmonger()
    certmonger_cmd_template = paths.CERTMONGER_COMMAND_TEMPLATE

    ca_path = cm.obj_if.find_ca_by_nickname(ca)
    if not ca_path:
        raise RuntimeError('{} CA not found'.format(ca))

    params = {
        'TRACK': True,
        'CERT_STORAGE': storage,
        'KEY_STORAGE': storage,
        'CERT_LOCATION': certfile,
        'KEY_LOCATION': keyfile,
        'CA': ca_path
    }
    if nickname:
        params['CERT_NICKNAME'] = nickname
        params['KEY_NICKNAME'] = nickname
    if pin:
        params['KEY_PIN'] = pin
    if pinfile:
        params['KEY_PIN_FILE'] = os.path.abspath(pinfile)
    if pre_command:
        if not os.path.isabs(pre_command):
            pre_command = certmonger_cmd_template % (pre_command)
        params['cert-presave-command'] = pre_command
    if post_command:
        if not os.path.isabs(post_command):
            post_command = certmonger_cmd_template % (post_command)
        params['cert-postsave-command'] = post_command
    if profile:
        params['ca-profile'] = profile

    result = cm.obj_if.add_request(params)
    try:
        if result[0]:
            request = _cm_dbus_object(cm.bus, cm, result[1], DBUS_CM_REQUEST_IF,
                                      DBUS_CM_IF, True)
        else:
            raise RuntimeError('add_request() returned False')
    except Exception as e:
        logger.error('Failed to add new request: %s', e)
        raise
    return request.prop_if.Get(DBUS_CM_REQUEST_IF, 'nickname')


def stop_tracking(secdir=None, request_id=None, nickname=None, certfile=None):
    """
    Stop tracking the current request using either the request_id or nickname.

    Returns True or False
    """
    if request_id is None and nickname is None and certfile is None:
        raise RuntimeError('One of request_id, nickname and certfile is'
                           ' required.')
    if secdir is not None and certfile is not None:
        raise RuntimeError("Can't specify both secdir and certfile.")

    criteria = dict()
    if secdir:
        criteria['cert-database'] = secdir
    if request_id:
        criteria['nickname'] = request_id
    if nickname:
        criteria['cert-nickname'] = nickname
    if certfile:
        criteria['cert-file'] = certfile
    try:
        request = _get_request(criteria)
    except RuntimeError as e:
        logger.error('Failed to get request: %s', e)
        raise
    if request:
        request.parent.obj_if.remove_request(request.path)


def modify(request_id, ca=None, profile=None, template_v2=None):
    update = {}
    if ca is not None:
        cm = _certmonger()
        update['CA'] = cm.obj_if.find_ca_by_nickname(ca)
    if profile is not None:
        update['template-profile'] = profile
    if template_v2 is not None:
        update['template-ms-certificate-template'] = template_v2

    if len(update) > 0:
        request = _get_request({'nickname': request_id})
        request.obj_if.modify(update)


def resubmit_request(
        request_id,
        ca=None,
        profile=None,
        template_v2=None,
        is_ca=False):
    """
    :param request_id: the certmonger numeric request ID
    :param ca: the nickname for the certmonger CA, e.g. IPA or SelfSign
    :param profile: the profile to use, e.g. SubCA.  For requests using the
                    Dogtag CA, this is the profile to use.  This also causes
                    the Microsoft certificate tempalte name extension to the
                    CSR (for telling AD CS what template to use).
    :param template_v2: Microsoft V2 template specifier extension value.
                        Format: <oid>:<major-version>[:<minor-version>]
    :param is_ca: boolean that if True adds the CA basic constraint
    """
    request = _get_request({'nickname': request_id})
    if request:
        update = {}
        if ca is not None:
            cm = _certmonger()
            update['CA'] = cm.obj_if.find_ca_by_nickname(ca)
        if profile is not None:
            update['template-profile'] = profile
        if template_v2 is not None:
            update['template-ms-certificate-template'] = template_v2
        if is_ca:
            update['template-is-ca'] = True
            update['template-ca-path-length'] = -1  # no path length

        if len(update) > 0:
            request.obj_if.modify(update)
        request.obj_if.resubmit()


def _find_IPA_ca():
    """
    Look through all the certmonger CA files to find the one that
    has id=IPA

    We can use find_request_value because the ca files have the
    same file format.
    """
    cm = _certmonger()
    ca_path = cm.obj_if.find_ca_by_nickname('IPA')
    return _cm_dbus_object(cm.bus, cm, ca_path, DBUS_CM_CA_IF, DBUS_CM_IF, True)


def add_principal_to_cas(principal):
    """
    If the hostname we were passed to use in ipa-client-install doesn't
    match the value of gethostname() then we need to append
    -k host/HOSTNAME@REALM to the ca helper defined for
    /usr/libexec/certmonger/ipa-submit.

    We also need to restore this on uninstall.
    """
    ca = _find_IPA_ca()
    if ca:
        ext_helper = ca.prop_if.Get(DBUS_CM_CA_IF, 'external-helper')
        if ext_helper and '-k' not in shlex.split(ext_helper):
            ext_helper = '%s -k %s' % (ext_helper.strip(), principal)
            ca.prop_if.Set(DBUS_CM_CA_IF, 'external-helper', ext_helper)


def remove_principal_from_cas():
    """
    Remove any -k principal options from the ipa_submit helper.
    """
    ca = _find_IPA_ca()
    if ca:
        ext_helper = ca.prop_if.Get(DBUS_CM_CA_IF, 'external-helper')
        if ext_helper and '-k' in shlex.split(ext_helper):
            ext_helper = shlex.split(ext_helper)[0]
            ca.prop_if.Set(DBUS_CM_CA_IF, 'external-helper', ext_helper)


def modify_ca_helper(ca_name, helper):
    """
    Modify certmonger CA helper.

    Applies the new helper and return the previous configuration.
    """
    bus = dbus.SystemBus()
    obj = bus.get_object('org.fedorahosted.certmonger',
                         '/org/fedorahosted/certmonger')
    iface = dbus.Interface(obj, 'org.fedorahosted.certmonger')
    path = iface.find_ca_by_nickname(ca_name)
    if not path:
        raise RuntimeError("{} is not configured".format(ca_name))
    else:
        ca_obj = bus.get_object('org.fedorahosted.certmonger', path)
        ca_iface = dbus.Interface(ca_obj,
                                  'org.freedesktop.DBus.Properties')
        old_helper = ca_iface.Get('org.fedorahosted.certmonger.ca',
                                  'external-helper')
        ca_iface.Set('org.fedorahosted.certmonger.ca',
                     'external-helper', helper,
                     # Give dogtag extra time to generate cert
                     timeout=CA_DBUS_TIMEOUT)
        return old_helper


def get_pin(token):
    """
    Dogtag stores its NSS pin in a file formatted as token:PIN.

    The caller is expected to handle any exceptions raised.
    """
    with open(paths.PKI_TOMCAT_PASSWORD_CONF, 'r') as f:
        for line in f:
            (tok, pin) = line.split('=', 1)
            if token == tok:
                return pin.strip()
    return None


def check_state(dirs):
    """
    Given a set of directories and nicknames verify that we are no longer
    tracking certificates.

    dirs is a list of directories to test for. We will return a tuple
    of nicknames for any tracked certificates found.

    This can only check for NSS-based certificates.
    """
    reqids = []
    for dir in dirs:
        reqids.extend(get_requests_for_dir(dir))

    return reqids


def wait_for_request(request_id, timeout=120):
    for _i in range(0, timeout, 5):
        state = get_request_value(request_id, 'status')
        logger.debug("certmonger request is in state %r", state)
        if state in ('CA_REJECTED', 'CA_UNREACHABLE', 'CA_UNCONFIGURED',
                     'NEED_GUIDANCE', 'NEED_CA', 'MONITORING'):
            break
        time.sleep(5)
    else:
        raise RuntimeError("request timed out")

    return state

if __name__ == '__main__':
    request_id = request_cert(paths.HTTPD_ALIAS_DIR,
                              "cn=tiger.example.com,O=IPA",
                              "HTTP/tiger.example.com@EXAMPLE.COM", "Test")
    csr = get_request_value(request_id, 'csr')
    print(csr)
    stop_tracking(request_id)



# Authors:
#   Jason Gerard DeRose <jderose@redhat.com>
#
# Copyright (C) 2009  Red Hat
# see file 'COPYING' for use and warranty information
#
# This program is free software; you can redistribute it and/or modify
# it under the terms of the GNU General Public License as published by
# the Free Software Foundation, either version 3 of the License, or
# (at your option) any later version.
#
# This program is distributed in the hope that it will be useful,
# but WITHOUT ANY WARRANTY; without even the implied warranty of
# MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
# GNU General Public License for more details.
#
# You should have received a copy of the GNU General Public License
# along with this program.  If not, see <http://www.gnu.org/licenses/>.

"""
Simple description of return values.
"""
import six

from ipalib.plugable import ReadOnly, lock
from ipalib.capabilities import client_has_capability
from ipalib.text import _
from ipalib.util import apirepr

if six.PY3:
    unicode = str

class Output(ReadOnly):
    """
    Simple description of a member in the return value ``dict``.

    This class controls both the type of object being returned by
    a command as well as how the output will be displayed.

    For example, this class defines two return results: an entry
    and a value.

    >>> from ipalib import crud, output
    >>> class user(crud.Update):
    ...
    ...     has_output = (
    ...         output.Entry('result'),
    ...         output.value,
    ...     )

    The order of the values in has_output controls the order of output.
    If you have values that you don't want to be printed then add
    ``'no_display'`` to flags.

    The difference between ``'no_display'`` and ``'no_output'`` is
    that ``'no_output'`` will prevent a Param value from being returned
    at all. ``'no_display'`` will cause the API to return a value, it
    simply won't be displayed to the user. This is so some things may
    be returned that while not interesting to us, but may be to others.

    >>> from ipalib import crud, output
    >>> myvalue = output.Output('myvalue', unicode,
    ...     'Do not print this value', flags=['no_display'],
    ... )
    >>> class user(crud.Update):
    ...
    ...     has_output = (
    ...         output.Entry('result'),
    ...         myvalue,
    ...     )
    """

    type = None
    validate = None
    doc = None
    flags = []

    def __init__(self, name, type=None, doc=None, flags=[]):
        self.name = name
        if type is not None:
            if not isinstance(type, tuple):
                type = (type,)
            self.type = type
        if doc is not None:
            self.doc = doc
        self.flags = flags
        lock(self)

    def __repr__(self):
        return '%s(%s)' % (
            self.__class__.__name__,
            ', '.join(self.__repr_iter())
        )

    def __repr_iter(self):
        yield repr(self.name)
        for key in ('type', 'doc', 'flags'):
            value = self.__dict__.get(key)
            if not value:
                continue
            if isinstance(value, tuple):
                value = apirepr(list(value))
            else:
                value = repr(value)
            yield '%s=%s' % (key, value)


class Entry(Output):
    type = dict
    doc = _('A dictionary representing an LDAP entry')


emsg = """%s.validate_output() => %s.validate():
  output[%r][%d]: need a %r; got a %r: %r"""

class ListOfEntries(Output):
    type = (list, tuple)
    doc = _('A list of LDAP entries')

    def validate(self, cmd, entries, version):
        assert isinstance(entries, self.type)
        for (i, entry) in enumerate(entries):
            if not isinstance(entry, dict):
                raise TypeError(emsg % (cmd.name, self.__class__.__name__,
                    self.name, i, dict, type(entry), entry)
                )

class PrimaryKey(Output):
    def validate(self, cmd, value, version):
        if client_has_capability(version, 'primary_key_types'):
            if hasattr(cmd, 'obj') and cmd.obj and cmd.obj.primary_key:
                types = cmd.obj.primary_key.allowed_types
            else:
                types = (unicode,)
            types = types + (type(None),)
        else:
            types = (unicode,)
        if not isinstance(value, types):
            raise TypeError(
                "%s.validate_output() => %s.validate():\n"
                "  output[%r]: need %r; got %r: %r" % (
                    cmd.name, self.__class__.__name__, self.name,
                    types[0], type(value), value))

class ListOfPrimaryKeys(Output):
    def validate(self, cmd, values, version):
        if client_has_capability(version, 'primary_key_types'):
            types = (tuple, list)
        else:
            types = (unicode,)
        if not isinstance(values, types):
            raise TypeError(
                "%s.validate_output() => %s.validate():\n"
                "  output[%r]: need %r; got %r: %r" % (
                    cmd.name, self.__class__.__name__, self.name,
                    types[0], type(values), values))

        if client_has_capability(version, 'primary_key_types'):
            if hasattr(cmd, 'obj') and cmd.obj and cmd.obj.primary_key:
                types = cmd.obj.primary_key.allowed_types
            else:
                types = (unicode,)
            for (i, value) in enumerate(values):
                if not isinstance(value, types):
                    raise TypeError(emsg % (
                        cmd.name, self.__class__.__name__, i, self.name,
                        types[0], type(value), value))


result = Output('result', doc=_('All commands should at least have a result'))

summary = Output('summary', (unicode, type(None)),
    _('User-friendly description of action performed')
)

value = PrimaryKey('value', None,
    _("The primary_key value of the entry, e.g. 'jdoe' for a user"),
    flags=['no_display'],
)

standard = (summary, result)

standard_entry = (
    summary,
    Entry('result'),
    value,
)

standard_list_of_entries = (
    summary,
    ListOfEntries('result'),
    Output('count', int, _('Number of entries returned')),
    Output('truncated', bool, _('True if not all results were returned')),
)

standard_delete = (
    summary,
    Output('result', dict, _('List of deletions that failed')),
    value,
)

standard_multi_delete = (
    summary,
    Output('result', dict, _('List of deletions that failed')),
    ListOfPrimaryKeys('value', flags=['no_display']),
)

standard_boolean = (
    summary,
    Output('result', bool, _('True means the operation was successful')),
    value,
)

standard_value = standard_boolean

simple_value = (
    summary,
    Output('result', bool, _('True means the operation was successful')),
    Output('value', unicode, flags=['no_display']),
)

# custom shim for commands like `trustconfig-show`,
# `automember-default-group-*` which put stuff into output['value'] despite not
# having primary key themselves. Designing commands like this is not a very
# good practice, so please do not use this for new code.
simple_entry = (
    summary,
    Entry('result'),
    Output('value', unicode, flags=['no_display']),
)



# Authors: John Dennis <jdennis@redhat.com>
#
# Copyright (C) 2012  Red Hat
# see file 'COPYING' for use and warranty information
#
# This program is free software; you can redistribute it and/or modify
# it under the terms of the GNU General Public License as published by
# the Free Software Foundation, either version 3 of the License, or
# (at your option) any later version.
#
# This program is distributed in the hope that it will be useful,
# but WITHOUT ANY WARRANTY; without even the implied warranty of
# MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
# GNU General Public License for more details.
#
# You should have received a copy of the GNU General Public License
# along with this program.  If not, see <http://www.gnu.org/licenses/>.

import time
import re

import six
import gssapi

from ipalib import errors

if six.PY3:
    unicode = str

#-------------------------------------------------------------------------------

# Kerberos error codes
KRB5_CC_NOTFOUND                = 2529639053 # Matching credential not found
KRB5_FCC_NOFILE                 = 2529639107 # No credentials cache found
KRB5KDC_ERR_C_PRINCIPAL_UNKNOWN = 2529638918  # client not found in Kerberos db
KRB5KDC_ERR_S_PRINCIPAL_UNKNOWN = 2529638919 # Server not found in Kerberos database
KRB5KRB_AP_ERR_TKT_EXPIRED      = 2529638944 # Ticket expired
KRB5_FCC_PERM                   = 2529639106 # Credentials cache permissions incorrect
KRB5_CC_FORMAT                  = 2529639111 # Bad format in credentials cache
KRB5_REALM_CANT_RESOLVE         = 2529639132 # Cannot resolve network address for KDC in requested realm

krb_ticket_expiration_threshold = 60*5 # number of seconds to accmodate clock skew
krb5_time_fmt = '%m/%d/%y %H:%M:%S'
ccache_name_re = re.compile(r'^((\w+):)?(.+)')

#-------------------------------------------------------------------------------

def krb5_parse_ccache(ccache_name):
    '''
    Given a Kerberos ccache name parse it into it's scheme and
    location components. Currently valid values for the scheme
    are:

      * FILE
      * MEMORY

    The scheme is always returned as upper case. If the scheme
    does not exist it defaults to FILE.

    :parameters:
      ccache_name
        The name of the Kerberos ccache.
    :returns:
      A two-tuple of (scheme, ccache)
    '''
    match = ccache_name_re.search(ccache_name)
    if match:
        scheme = match.group(2)
        location = match.group(3)
        if scheme is None:
            scheme = 'FILE'
        else:
            scheme = scheme.upper()

        return scheme, location
    else:
        raise ValueError('Invalid ccache name = "%s"' % ccache_name)

def krb5_unparse_ccache(scheme, name):
    return '%s:%s' % (scheme.upper(), name)


def krb5_format_service_principal_name(service, host, realm):
    '''

    Given a Kerberos service principal name, the host where the
    service is running and a Kerberos realm return the Kerberos V5
    service principal name.

    :parameters:
      service
        Service principal name.
      host
        The DNS name of the host where the service is located.
      realm
        The Kerberos realm the service exists in.
    :returns:
      Kerberos V5 service principal name.
    '''
    return '%s/%s@%s' % (service, host, realm)

def krb5_format_tgt_principal_name(realm):
    '''
    Given a Kerberos realm return the Kerberos V5 TGT name.

    :parameters:
      realm
        The Kerberos realm the TGT exists in.
    :returns:
      Kerberos V5 TGT name.
    '''
    return krb5_format_service_principal_name('krbtgt', realm, realm)

def krb5_format_time(timestamp):
    '''
    Given a UNIX timestamp format it into a string in the same
    manner the MIT Kerberos library does. Kerberos timestamps are
    always in local time.

    :parameters:
      timestamp
        Unix timestamp
    :returns:
      formated string
    '''
    return time.strftime(krb5_time_fmt, time.localtime(timestamp))

def get_credentials(name=None, ccache_name=None):
    '''
    Obtains GSSAPI credentials with given principal name from ccache. When no
    principal name specified, it retrieves the default one for given
    credentials cache.

    :parameters:
      name
        gssapi.Name object specifying principal or None for the default
      ccache_name
        string specifying Kerberos credentials cache name or None for the
        default
    :returns:
      gssapi.Credentials object
    '''
    store = None
    if ccache_name:
        store = {'ccache': ccache_name}
    try:
        return gssapi.Credentials(usage='initiate', name=name, store=store)
    except gssapi.exceptions.GSSError as e:
        if e.min_code == KRB5_FCC_NOFILE:  # pylint: disable=no-member
            raise ValueError('"%s", ccache="%s"' % (e, ccache_name))
        raise

def get_principal(ccache_name=None):
    '''
    Gets default principal name from given credentials cache.

    :parameters:
      ccache_name
        string specifying Kerberos credentials cache name or None for the
        default
    :returns:
      Default principal name as string
    :raises:
      errors.CCacheError if the principal cannot be retrieved from given
      ccache
    '''
    try:
        creds = get_credentials(ccache_name=ccache_name)
        return unicode(creds.name)
    except gssapi.exceptions.GSSError as e:
        raise errors.CCacheError(message=unicode(e))

def get_credentials_if_valid(name=None, ccache_name=None):
    '''
    Obtains GSSAPI credentials with principal name from ccache. When no
    principal name specified, it retrieves the default one for given
    credentials cache. When the credentials cannot be retrieved or aren't valid
    it returns None.

    :parameters:
      name
        gssapi.Name object specifying principal or None for the default
      ccache_name
        string specifying Kerberos credentials cache name or None for the
        default
    :returns:
      gssapi.Credentials object or None if valid credentials weren't found
    '''

    try:
        creds = get_credentials(name=name, ccache_name=ccache_name)
        if creds.lifetime > 0:
            return creds
        return None
    except gssapi.exceptions.ExpiredCredentialsError:
        return None
    except gssapi.exceptions.GSSError:
        return None
